
/******************************************************************************
* MODULE     : real.hpp
* DESCRIPTION: Effective real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_REAL_HPP
#define __MML_REAL_HPP
#include "dyadic.hpp"
#include "interval.hpp"
namespace mml {

#define MAX_CANCEL_TEST  256
#define MAX_CANCEL_ERROR 4096

class real_rep {
public:
  const dyadic wt; // weight of real number as an expression
  interval best;   // the best currently available approximation
                   // FIXME: best should be protected
public:
  inline real_rep (const dyadic& wt2):
    wt (wt2), best (interval::fuzzy) {}
  inline real_rep (const dyadic& wt2, const interval& best2):
    wt (wt2), best (best2) {}
  virtual inline ~real_rep () {}

  virtual void print (ostream& out) = 0;
  virtual dyadic get_bound () = 0;
  virtual interval approx (const dyadic& err) = 0;

  mp_prec_t linear_algo (const dyadic& err);
  mp_prec_t mult_algo (const dyadic& err);
  bool normalize (mp_prec_t max_cancel= MAX_CANCEL_TEST,
		  const dyadic& rel_err= error (1, -MIN_SECURITY));
};

class real {
INDIRECT_PROTO (real, real_rep)
public:
  real (signed int i);
  real (unsigned int i);
  real (signed short int i);
  real (unsigned short int i);
  real (signed long int i);
  real (unsigned long int i);
};
INDIRECT_IMPL (real, real_rep)

inline dyadic weight (const real& x1) {
  return x1->wt; }
inline dyadic weight (const real& x1, const real& x2) {
  return add_up (x1->wt, x2->wt); }

inline ostream& operator << (ostream& out, const real& x) {
  x->print (out); return out; }
inline dyadic bound (const real& x) {
  if (is_normal (x->best)) return bound (x->best);
  else return x->get_bound (); }

real operator - (const real& x);
real operator + (const real& x, const real& y);
real operator - (const real& x, const real& y);
real operator * (const real& x, const real& y);
real operator / (const real& x, const real& y);
real operator << (const real& x, mp_exp_t shift);
real operator >> (const real& x, mp_exp_t shift);
real abs (const real& x);

inline real& operator += (real& x, const real& y) { return x= x+y; }
inline real& operator -= (real& x, const real& y) { return x= x-y; }
inline real& operator *= (real& x, const real& y) { return x= x*y; }
inline real& operator /= (real& x, const real& y) { return x= x/y; }
inline real& operator <<= (real& x, mp_exp_t sh) { return x= x<<sh; }
inline real& operator >>= (real& x, mp_exp_t sh) { return x= x>>sh; }

int sign (const real& x, int max_cancel= MAX_CANCEL_TEST);
inline bool operator == (const real& x, const real& y) {
  return sign (x-y) == 0; }
inline bool operator != (const real& x, const real& y) {
  return sign (x-y) != 0; }
inline bool operator < (const real& x, const real& y) {
  return sign (x-y) < 0; }
inline bool operator > (const real& x, const real& y) {
  return sign (x-y) > 0; }
inline bool operator <= (const real& x, const real& y) {
  return sign (x-y) <= 0; }
inline bool operator >= (const real& x, const real& y) {
  return sign (x-y) >= 0; }

}
#endif // __MML_REAL_HPP
