
/******************************************************************************
* MODULE     : real_mul.cpp
* DESCRIPTION: Multiplication of real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "real.hpp"
namespace mml {

class mul_real_rep: public real_rep {
private:
  real x, y;
public:
  inline mul_real_rep (real x2, real y2):
    real_rep (weight (x2, y2)), x (x2), y (y2) {}
  void print (ostream& out) { out << "(* " << x << " " << y << ")"; }
  dyadic get_bound () { return mul_up (bound (x), bound (y)); }
  interval approx (const dyadic& err);
};

interval
mul_real_rep::approx (const dyadic& err) {
  if (wider (best, err)) {
    if (has_huge_prec (best)) {
      dyadic rel_r= err / magnitude (best);
      dyadic rel_x= spread (rel_r, x->wt, wt);
      dyadic rel_y= spread (rel_r, y->wt, wt);
      interval xa= x->approx (magnitude (x->best) * rel_x);
      interval ya= y->approx (magnitude (y->best) * rel_y);
      mp_prec_t new_prec= mult_algo (err);
      best= reduce (xa, new_prec) * reduce (ya, new_prec);
    }
    else {
      interval xa= x->approx (lp_infinity);
      interval ya= y->approx (lp_infinity);
      if (is_epsilon (xa)) return xa;
      if (is_epsilon (ya)) return ya;
      dyadic bound_x= bound (x);
      dyadic bound_y= bound (y);
      dyadic bound_r= bound_x * bound_y;
      dyadic rel_r  = min (err / bound_r, 1);
      if ((rel_r << (DEFAULT_LOW<<1)) < 1) {
	// Since we lack information for making good a priori error estimates,
	// we first compute an approximation at a more modest precision and
	// then check whether we may use the huge precision error estimates.
	dyadic modest_err= sqrt (rel_r) * bound_r;
	(void) approx (modest_err);
	if (has_huge_prec (best)) return approx (err);
      }
      dyadic rel_x= spread (rel_r, x->wt, wt);
      dyadic rel_y= spread (rel_r, y->wt, wt);
      xa= x->approx (rel_x * bound_x);
      ya= y->approx (rel_y * bound_y);
      mp_prec_t new_prec= mult_algo (err);
      best= reduce (xa, new_prec) * reduce (ya, new_prec);
    }
    if (wider (best, err))
      fatal_error ("Bad a priori error estimation",
		   "mul_real_rep::approx", "real_mul.cpp");
  }
  return best;
}

real
operator * (const real& x, const real& y) {
  return new mul_real_rep (x, y);
}

}
